<?php
/**
 * EPL eSync Pro(visioning): plugin
 *
 * @link www.egroupware.org
 * @author Ralf Becker <rb(at)egroupware.org>
 * @copyright (c) 2012-16 by Ralf Becker <rb(at)egroupware.org>
 * @package esyncpro
 * @subpackage policy
 * @version $Id$
 */

use EGroupware\Api;

/**
 * eSync plugin
 *
 * @todo re-implement remote wipe for zpush2 and check it
 * @todo re-implement device specific logging
 * @todo re-implement deleting of profiles
 * @todo device check device multiple users or changed user
 */
class esyncpro_zpush
{
	/**
	 * @var activesync_backend
	 */
	private $backend;

	/**
	 * Instance of device class
	 *
	 * @var esyncpro_device
	 */
	private $device;

	/**
	 * Instance of policy class
	 *
	 * @var esyncpro_policy
	 */
	private $policy;

	/**
	 * Constructor
	 *
	 * @param activesync_backend $backend
	 */
	public function __construct(activesync_backend $backend)
	{
		$this->backend = $backend;
	}

	/**
	 * Return a device wipe status
	 *
	 * Z-Push will send remote wipe request to client, if returned status is SYNC_PROVISION_RWSTATUS_PENDING or _WIPED
	 *
	 * Plugins either return array() to NOT change standard response or array('response' => value)
	 *
	 * @param string $user
	 * @param string $pass
	 * @param string $devid
	 * @return array('response' => SYNC_PROVISION_RWSTATUS_NA, SYNC_PROVISION_RWSTATUS_OK, SYNC_PROVISION_RWSTATUS_PENDING, SYNC_PROVISION_RWSTATUS_WIPED
	 */
	function getDeviceRWStatus($user, $pass, $devid)
	{
		unset($user, $pass);	// not used, but required by function signature

		$device = $this->read_init_device($devid);
		// triggers a remote wipe, if status is _PENDING or _WIPED
		$response = array('response' => $device['dev_rw_status']);

		//error_log(__METHOD__."('$user', '$pass', '$devid') returning ".array2string($response['response']));
		return $response;
	}

	/**
	 * Set a new rw status for the device
	 *
	 * Z-Push call this with SYNC_PROVISION_RWSTATUS_WIPED, after sending remote wipe command
	 *
	 * Plugins either return array() to NOT change standard response or array('response' => value)
	 *
	 * @param string $user
	 * @param string $pass
	 * @param string $devid
	 * @param int $status SYNC_PROVISION_RWSTATUS_OK, SYNC_PROVISION_RWSTATUS_PENDING, SYNC_PROVISION_RWSTATUS_WIPED
	 * @return boolean seems not to be used in Z-Push
	 */
	function setDeviceRWStatus($user, $pass, $devid, $status)
	{
		unset($user, $pass);	// not used, but required by function signature

		$response = array();
		$device = $this->read_init_device($devid);

		if ($device['dev_rw_status'] != $status)
		{
			$device['dev_rw_status'] = $status;
			if ($status == SYNC_PROVISION_RWSTATUS_WIPED) $device['dev_wiped'] = $this->now;
			$this->device->save($device);
		}
		return $response;
	}

	/**
	 * Checks if the sent policykey matches the latest policykey on the server
	 *
	 * Plugins either return array() to NOT change standard response or array('response' => value)
	 *
	 * @param string $policykey
	 * @param string $devid
	 *
	 * @return boolean true: policy ok, false: provisioning necessary
	 */
	function CheckPolicy($policykey, $devid)
	{
		$response = true;
		$device = $this->read_init_device($devid);

		// if current policy was applied, check if policy itself got updated
		if ($device['dev_policy_applied'])
		{
			if (!isset($this->policy)) $this->policy = new esyncpro_policy();
			if ($device['policy_id'] && ($this->policy->data['policy_id'] == $device['policy_id'] ||
				($this->policy->read(array('policy_id' => $device['policy_id'])))))
			{
				$policy = $this->policy->data;
			}
			else
			{
				$policy = $this->policy->standard(false);
			}
			if ($policy['policy_modified'] &&
					Api\DateTime::to($policy['policy_modified'], 'ts') > Api\DateTime::to($device['dev_policy_applied'], 'ts') ||
				// policy update interval set and due
				$policy['policy_update'] &&
					Api\DateTime::to('now', 'ts')-egw_time::to($device['dev_policy_applied'], 'ts') > 3600*$policy['policy_update'] ||
				$device['dev_rw_status'] == SYNC_PROVISION_RWSTATUS_PENDING ||
				$device['dev_rw_status'] == SYNC_PROVISION_RWSTATUS_WIPED)
			{
				// policy changed after it was applied --> return mismatch, to force provisioning
				$response = false;
			}
		}
		else
		{
			// current policy never applied --> return mismatch, to force provisioning
			$response = false;
		}
		ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."('$policykey', '$devid') device[policy_id]=$device[policy_id], policy_id=$policy[policy_id], device[policy_applied]=$device[dev_policy_applied], policy_modified=$policy[policy_modified] returning ".array2string($response));
		return $response;
	}

	/**
	 * Checks if policy of device allows loose provisioning
	 *
	 * Plugins either return array() to NOT change standard response or array('response' => value)
	 *
	 * @param string $policykey
	 * @param string $devid
	 *
	 * @return boolean
	 */
	function LooseProvisioning($devid)
	{
		$device = $this->read_init_device($devid);

		// read device or standard policy
		if (!isset($this->policy)) $this->policy = new esyncpro_policy();
		if ($device['policy_id'] && ($this->policy->data['policy_id'] == $device['policy_id'] ||
			($this->policy->read(array('policy_id' => $device['policy_id'])))))
		{
			$policy = $this->policy->data;
		}
		else
		{
			$policy = $this->policy->standard(false);
		}
		// if current policy was applied, check if policy itself got updated
		$response = array('response' => (boolean)$policy['policy_loose']);

		//ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."('$devid') returning ".array2string($response));
		return $response;
	}

	/**
	 * Settings Support, storing information in device list
	 *
	 * @param SyncObject $settings
	 */
	public function Settings(SyncObject $settings)
	{
		//ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."(".array2string($settings).")");
		unset($settings);	// not used, but required by function signature

		return array();
	}

	/**
	 * Writes ta state to for a key and counter
	 *
	 * @param StateObject $state
	 * @param string    $devid              the device id
	 * @param string    $type               the state type
	 * @param string    $key                (opt)
	 * @param int       $counter            (opt)
	 */
	public function SetState($state, $devid, $type, $key = false, $counter = false)
	{
		//ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."(..., '$devid', '$type', '$key', $counter)");

		if (!Request::IsValidDeviceID()) return;

		switch($type)
		{
			case 'devicedata':
				$device = $state->devices[Request::GetAuthUser()];
				ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."(..., '$devid', '$type', '$key', $counter) device=".array2string($device).', prev. policykey='.$this->device->data['policykey'].', this->device->data='.array2string($this->device->data).', Request::GetPolicyKey()='.Request::GetPolicyKey());
				if ($device instanceof ASDevice)
				{
					static $name2col = array(
						'devicemodel'          => 'dev_model',
						'deviceimei'           => 'dev_imei',
						'devicefriendlyname'   => 'dev_name',
						'deviceos'             => 'dev_os',
						'deviceoslanguage'     => 'dev_lang',
						'devicephonenumber'    => 'dev_phoneno',
						'devicemobileoperator' => 'dev_operator',
						'useragent'            => 'dev_useragent',
						//'enableoutboundsms'
					);
					$data = $this->read_init_device($devid);

					// store settings deviceinformation
					foreach($name2col as $name => $col)
					{
						if (isset($device->$name)) $data[$col] = $device->$name;
					}

					// check if policy was just confirmed by client
					if (!$this->device->data['dev_policy_applied'] && Request::WasPolicyKeySent() && Request::GetPolicyKey() == $this->device->data['policykey'])
					{
						ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."() policykey={$this->device->data['policykey']} applied");
						$data['dev_policy_applied'] = time();
					}
					ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."() device->save(".array2string($data).")");
					//error_log(__METHOD__.'('.array2string($settings).", '$devid') saving device-data ".array2string($data));
					$this->device->save($data);
				}
				break;
		}
		return array();
	}

	/**
	 * Writes ta state to for a key and counter
	 *
	 * @param StateObject $state
	 * @param string    $devid              the device id
	 * @param string    $type               the state type
	 * @param string    $key                (opt)
	 * @param int       $counter            (opt)
	 *
	 * @access public
	 * @return boolean
	 * @throws StateInvalidException
	 */
	public function GetState($state, $devid, $type, $key = false, $counter = false)
	{
		//ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."(..., '$devid', '$type', '$key', $counter)");

		switch($type)
		{
			case 'devicedata':
				$device = $state->devices[Request::GetAuthUser()];
				if ($device instanceof ASDevice)
				{
					$this->read_init_device($devid);

					// we do NOT check policy, if z-push and device run their provisioning
					if (Request::GetCommand() != 'Provision' && !$this->CheckPolicy($device->policykey, $devid))
					{
						$device->policykey = DeviceManager::GenerateProvisioningPolicyKey();
						$this->device->data['dev_policy_applied'] = null;
						ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."(..., '$devid', '$type', '$key', $counter) force provisioning");
					}
					$this->device->data['policykey'] = $device->policykey;
					$this->device->save();
					$device->policies = $this->getProvision($devid, Request::GetAuthUser());
					ZLog::Write(LOGLEVEL_DEBUG, __METHOD__."(..., '$devid', '$type', '$key', $counter) policykey=$device->policykey");
				}
		}
		return array();
	}

	/**
	 * Get provisioning data for a given device and user
	 *
	 * @param string $devid
	 * @param string $user
	 * @return array
	 */
	public function getProvision($devid, $user)
	{
		$device = $this->read_init_device($devid);

		if (!isset($this->policy)) $this->policy = new esyncpro_policy();
		if ($device['policy_id'] && ($this->policy->data['policy_id'] == $device['policy_id'] ||
			($this->policy->read(array('policy_id' => $device['policy_id'])))))
		{
			$policy = $this->policy->data;
		}
		else
		{
			$policy = $this->policy->standard(false);
		}
		$device['account_id'] = $GLOBALS['egw_info']['user']['account_lid'] == $user ?
			$GLOBALS['egw_info']['user']['account_id'] :
			$GLOBALS['egw']->accounts->name2id($user);

		// save updated device data
		$device['policy_id'] = $policy['policy_id'];
		$device['dev_policy_applied'] = $device['dev_last_sync'] = time();
		//error_log(__METHOD__."('$devid', '$user') saving device-data ".array2string($device));
		if ($devid != 'validate') $this->device->save($device);

		$provision = array();
		foreach(array(
			'DevicePasswordEnabled' => 'policy_pw_required',	// '0'
			'AlphanumericDevicePasswordRequired' => 'policy_pw_alphanum',	// '0',
			'PasswordRecoveryEnabled' => 'policy_pw_recovery',	// '1',
			'MinDevicePasswordLength' => 'policy_pw_length',	// '1',
			'MaxDevicePasswordFailedAttempts' => 'policy_pw_max_failed',	// '5',
			'AllowSimpleDevicePassword' => 'policy_pw_simple',	// '1',
			'DevicePasswordExpiration' => 'policy_pw_expiration',	// '',
			'DevicePasswordHistory' => 'policy_pw_history',	// '0',
			'DeviceEncryptionEnabled' => 'policy_dev_encryption',	// '0',
			'AttachmentsEnabled' => 'policy_attachments',	// '1',
			'MaxInactivityTimeDeviceLock' => 'policy_max_inactive_lock',	// '9999',
			'MaxAttachmentSize' => 'policy_max_attachment_size',	// '5000000',
			'AllowStorageCard' => 'policy_allow_card',	// '1',
			'AllowCamera' => 'policy_allow_camera',	// '1',
			'RequireDeviceEncryption' => 'policy_dev_encryption',	// '0',
			// not (yet) in EGw policy
			//'AllowUnsignedApplications' => '1',
			//'AllowUnsignedInstallationPackages' => '1',
			'MinDevicePasswordComplexCharacters' => 'policy_pw_complexity',	// '3',
			'AllowWiFi' => 'policy_allow_wifi',	// '1',
			'AllowTextMessaging' => 'policy_allow_sms',	// '1',
			'AllowPOPIMAPEmail' => 'policy_allow_popimap',	// '1',
			'AllowBluetooth' => 'policy_allow_bluetooth',	// '2',
			'AllowIrDA' => 'policy_allow_irda',	// '1',
			'RequireManualSyncWhenRoaming' => 'policy_roaming_manual',	// '0', // Set to one in case you'd like to save money...
			'AllowDesktopSync' => 'policy_allow_desktop_sync',	// '1',
			'MaxCalendarAgeFilter' => 'policy_max_cal_age',	// '0',
			'AllowHTMLEmail' => 'policy_allow_html',	// '1',
			'MaxEmailAgeFilter' => 'policy_max_email_age',	// '0',
			'MaxEmailBodyTruncationSize' => 'policy_max_email_size',	// '5000000',
			'MaxHTMLBodyTruncationSize' => 'policy_max_email_size',		// '5000000',
			//'RequireSignedSMIMEMessages' => '0',
			//'RequireEncryptedSMIMEMessages' => '0',
			//'RequireSignedSMIMEAlgorithm' => '0',
			//'RequireEncryptedSMIMEAlgorithm' => '0',
			//'AllowSMIMEEncryptionAlgorithmNegotiation' => '2', // was 1
			//'AllowSMIMESoftCerts' => '1',
			'AllowBrowser' => 'policy_allow_browser',	// '1',
			'AllowConsumerEmail' => 'policy_allow_consumer_email',	// '1',
			'AllowRemoteDesktop' => 'policy_allow_rdesktop',	// '1',
			'AllowInternetSharing' => 'policy_allow_inetsharing',	// '1',
			//'UnapprovedInROMApplicationList' => '',
			//'ApprovedApplicationList' => '',
		) as $as_name => $col)
		{
			switch($as_name)
			{
				case 'DevicePasswordExpiration':	// 0 is returned as ''
					if (!$policy[$col]) $policy[$col] = '';
					break;

				case 'MaxInactivityTimeDeviceLock':	// convert to seconds, as we store minutes
					$policy[$col] *= 60;
					break;

				case 'MaxAttachmentSize':
					if (!isset($policy[$as_name])) continue 2;	// not set use AS default (switch itself is 1!)
					// fall through
				case 'MaxEmailBodyTruncationSize':
				case 'MaxHTMLBodyTruncationSize':	// convert to bytes, as we store kb
					if ($policy[$col] > 0)
					{
						$policy[$col] *= 1024;
					}
					break;
			}
			// return boolean as 0 or 1
			if (is_bool($policy[$col])) $policy[$col] = (int)$policy[$col];

			// return everything as string
			$provision['Provision:'.$as_name] = (string)$policy[$col];
		}
		//error_log(__METHOD__."('$devid', '$user') returning ".array2string($provision));
		return $provision;
	}

	/**
	 * Log into EGroupware
	 *
	 * Implemented to log command and date of each AS request
	 *
	 * @param string $username
	 * @param string $domain
	 * @param string $password
	 * @return array not used, if no array you get a warning
	 */
	public function Logon($username, $domain, $password)
	{
		unset($username, $domain, $password);	// not used, but required by function signature

		$devid = Request::GetDeviceID();
		$cmd = Request::GetCommand();

		if ($devid && $devid != 'validate' && ($device = $this->read_init_device($devid)))
		{
			$device['dev_last_sync'] = $this->device->now;
			$device['dev_last_cmd']  = $cmd;
			$this->device->save($device);
		}
		return array();
	}

	/**
	 * Read device information and return it, if none stored call init with given devid
	 *
	 * Init $this->device, if not yet done
	 *
	 * @param string $devid
	 * @return array
	 */
	private function read_init_device($devid)
	{
		if (!isset($this->device)) $this->device = new esyncpro_device();

		if ($this->device->data['dev_uid'] !== $devid)
		{
			if (!($device = $this->device->read(array('dev_uid' => $devid))))
			{
				$device = $this->device->init(array('dev_uid' => $devid));
			}
		}
		else
		{
			$device = $this->device->data;
		}
		// set global available information
		if (!isset($device['account_id'])) $device['account_id'] = $GLOBALS['egw_info']['user']['account_id'];
		$protocolversion = Request::GetProtocolVersion();
		if (!empty($protocolversion)) $device['dev_as_version'] = $protocolversion;
		$useragent = Request::GetUserAgent();
		if (!empty($useragent)) $device['dev_useragent'] = $useragent;
		$devtype = Request::GetDeviceType();
		if (!empty($devtype)) $device['dev_type'] = $devtype;

		return $device;
	}

	/**
	 * Hook into esync plugins, without user having access to provisioning
	 *
	 * @param string|array $location
	 * @return string|array
	 */
	public static function esync_extra_apps($location)
	{
		unset($location);	// not used, but required by function signature

		return 'esyncpro';
	}
}